//===========================================================================
// Copyright  2000 Thin Air Enterprises and Robert Dunn.  All rights reserved.
// Free for non-commercial use.  Commercial use requires license agreement.
// See http://home.att.net/~robertdunn/Yacs.html for the most current version.
//===========================================================================
//---------------------------------------------------------------------------
// LogoMain.cpp - main form for regular expression search demonstration.
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "About.h"
#include "LogoMain.h"
#include "LogoStrs.h"
#include "MAPI.hpp"
#include "FindDlg.h"
#include "TaeRegExp.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TLogoAppForm *LogoAppForm;
//---------------------------------------------------------------------------
__fastcall TLogoAppForm::TLogoAppForm(TComponent *Owner)
	: TForm(Owner)
{
		FileNew1->Execute(); /* sets the default file name and clears the RichEdit Control */
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::FileNew1Execute(TObject *Sender)
{
	FFileName = LoadStr(sUntitled);
	RichEdit1->Lines->Clear();
	RichEdit1->Modified = false;
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::FileOpen1Execute(TObject *Sender)
{
	if (OpenDialog->Execute())
	{
		RichEdit1->Lines->LoadFromFile(OpenDialog->FileName);
		FFileName = OpenDialog->FileName;
		RichEdit1->SetFocus();
		RichEdit1->Modified = false;
		RichEdit1->ReadOnly = OpenDialog->Options.Contains(ofReadOnly);
	}
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::FileSave1Execute(TObject *Sender)
{
	if (FFileName == LoadStr(sUntitled))
		FileSaveAs1Execute(Sender);
	else
	{
		RichEdit1->Lines->SaveToFile(FFileName);
		RichEdit1->Modified = false;
	}
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::FileSaveAs1Execute(TObject *Sender)
{
	String str;
	TVarRec vrs[1];

	if (SaveDialog->Execute())
	{
		if (FileExists(SaveDialog->FileName))
		{
			str = FmtLoadStr(sOverwrite, OPENARRAY(TVarRec, (SaveDialog->FileName)));
			if (MessageDlg(str, mtConfirmation, TMsgDlgButtons() << mbYes << mbNo <<
				mbCancel, 0) != IDYES)
				return;
		}
		RichEdit1->Lines->SaveToFile(SaveDialog->FileName);
		FFileName = SaveDialog->FileName;
		RichEdit1->Modified = false;
	}
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::FileSend1Execute(TObject *Sender)
{
	TMapiMessage MapiMessage;
	Cardinal MError;

	MapiMessage.ulReserved = 0;
	MapiMessage.lpszSubject = NULL;
	MapiMessage.lpszNoteText = RichEdit1->Lines->Text.c_str();
	MapiMessage.lpszMessageType = NULL;
	MapiMessage.lpszDateReceived = NULL;
	MapiMessage.lpszConversationID = NULL;
	MapiMessage.flFlags = 0;
	MapiMessage.lpOriginator = NULL;
	MapiMessage.nRecipCount = 0;
	MapiMessage.lpRecips = NULL;
	MapiMessage.nFileCount = 0;
	MapiMessage.lpFiles = NULL;

	MError = MapiSendMail(0, reinterpret_cast<unsigned int>(Application->Handle),
				 MapiMessage, MAPI_DIALOG | MAPI_LOGON_UI | MAPI_NEW_SESSION, 0);
	if (MError)
		MessageDlg(LoadStr(sSendError), mtError, TMsgDlgButtons() << mbOK, 0);
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::FileExit1Execute(TObject *Sender)
{
	Close();
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::HelpAbout1Execute(TObject *Sender)
{
	AboutBox->ShowModal();
}
//---------------------------------------------------------------------------
// changes to the standard IDE-generated Logo Application begin here...
//---------------------------------------------------------------------------
// instantiate a TTaeRegExp object and a global AnsiString for the expression
//
TTaeRegExp* reObj = new TTaeRegExp();
AnsiString RegExp;

//---------------------------------------------------------------------------
// EnableFindAgain() - enable the repeat search menu and button
//
void __fastcall TLogoAppForm::EnableFindAgain(bool enable)
{
	SearchNext1->Enabled = enable;
	SearchPrev1->Enabled = enable;
}
//---------------------------------------------------------------------------
// need to know if a change occurs while searching -- the mere fact that
// the Rich Edit's event property is set to this event handler is used
// to determine that no change has occurred...
//
void __fastcall TLogoAppForm::RichEdit1Change(TObject *Sender)
{
	// unlink this handler
	RichEdit1->OnChange = 0;
}
//---------------------------------------------------------------------------
// MatchFailed() - handle a search failure; this is not called for a simple
// "no match condition"
//
void __fastcall TLogoAppForm::MatchFailed(void)
{
	// else if error
	EnableFindAgain(false);
	Application->MessageBox(reObj->LastErrMsg.c_str(), "Unexpected Error", MB_OK);
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::SearchFind1Execute(TObject *Sender)
{
	// set compile flag checkboxes from reObj
	FindForm->AnchoredCB->Checked = reObj->Anchored;
	FindForm->CaselessCB->Checked = reObj->Caseless;
	FindForm->DollarEndOnlyCB->Checked = reObj->DollarEndOnly;
	FindForm->DotAllCB->Checked = reObj->DotAll;
	FindForm->ExtendedCB->Checked = reObj->Extended;
	FindForm->ExtraCB->Checked = reObj->Extra;
	FindForm->MultiLineCB->Checked = reObj->MultiLine;
	FindForm->UngreedyCB->Checked = reObj->Ungreedy;

	// display find dialog
	TModalResult mrResult = FindForm->ShowModal();
	if (mrResult == mrCancel) return;

	// disable find next
	EnableFindAgain(false);

	// empty the search string
	RegExp = "";

	// get the regular expression (note:  we could use FindForm->RegExpMemo->Text
	// but it would return CRLFs at linebreaks; the below eliminates CRLFs)
	for (int i = 0; i < FindForm->RegExpMemo->Lines->Count; i++) {
		if (FindForm->RegExpMemo->Lines->Strings[i].Length() == 0) continue;
		RegExp += FindForm->RegExpMemo->Lines->Strings[i];
		}

	// get compile flags from FindForm
	reObj->Anchored = FindForm->AnchoredCB->Checked;
	reObj->Caseless = FindForm->CaselessCB->Checked;
	reObj->DollarEndOnly = FindForm->DollarEndOnlyCB->Checked;
	reObj->DotAll = FindForm->DotAllCB->Checked;
	reObj->Extended = FindForm->ExtendedCB->Checked;
	reObj->Extra = FindForm->ExtraCB->Checked;
	reObj->MultiLine = FindForm->MultiLineCB->Checked;
	reObj->Ungreedy = FindForm->UngreedyCB->Checked;

	// compile the regular expression
	bool result = reObj->Compile(RegExp);

	// display message and exit on error
	if (!result) {
		AnsiString errMsg = "An error was detected when parsing the Regular "
			"Expression.\n\r\n\rThe full expression was: \"" +
			RegExp + "\".\n\r\n\r"
			"The following portion of the expression was successfully "
			"evaluated: \"";
		errMsg += RegExp.SubString(1, reObj->LastErrOffset);
		errMsg += "\".\n\r\n\rThe error was recognized beginning with the "
			"following portion of the expression:  \"";
		errMsg += RegExp.SubString(reObj->LastErrOffset + 1, 9999);
		errMsg += "\".\n\r\n\rThe error message returned by the Regular Expression "
			"parser was: \"";
		errMsg += reObj->LastErrMsg;
		errMsg += "\".";
		Application->MessageBox(errMsg.c_str(), "Error Evaluating Regular Expression",
			MB_OK);
		return;
		}

	// we now have a functional compiled expression; enable repeats (we allow
	// repeating a search for any successfully compiled pattern regardless
	// of whether the RE contains any text)
	EnableFindAgain(true);

	// search for the expression (forwards = mrYes, backwards = mrNo)
	if (mrResult == mrYes) SearchNext1Execute(this);
	else SearchPrev1Execute(this);
}
//---------------------------------------------------------------------------
// FormShow() - attempt to load the "RegExpSyntax.rtf" document; on failure,
// display alternate text
//
void __fastcall TLogoAppForm::FormShow(TObject *Sender)
{
	try {
		RichEdit1->Lines->LoadFromFile("RegExpSyntax.rtf");
		}
	catch (...) {
		RichEdit1->SelText = "For information about using regular expressions, "
			"see the \"RegExpSyntax.rtf\" document distributed with this code or "
			"see the Borland C++ Builder Help topic \"PERL-compatible regular "
			"expressions\".";
		}
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::SearchNext1Execute(TObject *Sender)
{
	BuildMatchTable();
	MoveNext();
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::SearchPrev1Execute(TObject *Sender)
{
	BuildMatchTable();
	MovePrev();
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::BuildMatchTable(void)
{
	// build match table only if needed
	if (RichEdit1->OnChange) return;

	// scan for matches
	bool result = reObj->Scan(RichEdit1);

	// on error, display message
	if (!result) {
		AnsiString errMsg = "An unexpected error was encountered while "
			"building a table of matches.  The specific error message "
			"was:\n\r\n\r";
		errMsg += reObj->LastErrMsg;
		Application->MessageBox(errMsg.c_str(), "Error Building Match Table",
			MB_OK);
			return;
		}
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::MoveNext(void)
{
	// determine where the caret is currently located; locate index past
	// current selection
	int selPos = RichEdit1->SelStart + RichEdit1->SelLength;
	unsigned int ndx = reObj->FindNext(selPos);

	// if failure, beep
	if (ndx == (unsigned int) -1) {
		::MessageBeep(0);
		return;
		}

	// select match
	SelectText(ndx);
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::MovePrev(void)
{
	// determine where the caret is currently located; locate index before
	// current selection
	int selPos = RichEdit1->SelStart;
	unsigned int ndx = reObj->FindPrior(selPos);

	// if failure, beep
	if (ndx == (unsigned int) -1) {
		::MessageBeep(0);
		return;
		}

	// select match
	SelectText(ndx);
}
//---------------------------------------------------------------------------
void __fastcall TLogoAppForm::SelectText(int index)
{
	RichEdit1->SelStart = reObj->MatchStart[index];
	RichEdit1->SelLength = reObj->MatchLength[index];
}
//---------------------------------------------------------------------------
// end of LogoMain.cpp
//---------------------------------------------------------------------------
//===========================================================================
// Copyright  2000 Thin Air Enterprises and Robert Dunn.  All rights reserved.
// Free for non-commercial use.  Commercial use requires license agreement.
// See http://home.att.net/~robertdunn/Yacs.html for the most current version.
//===========================================================================


